@@ -1,16 +1,17 @@
-import argparse
 import collections
 import configparser
 import logging
 import os
+import sys
 from abc import ABCMeta
-from argparse import ArgumentParser
+from argparse import Action, ArgumentParser
 from enum import Enum
+from functools import partial
 from pathlib import Path
-from types import MappingProxyType
+from types import MappingProxyType, SimpleNamespace
 from typing import (
-    Any, Dict, Iterable, Mapping, NamedTuple, Optional, Sequence, Tuple, Type,
-    TypeVar, Union,
+    Any, Callable, Dict, Iterable, Mapping, MutableMapping, Optional, Sequence,
+    Set, Tuple, Type, TypeVar, Union,
 )
 
 
@@ -18,38 +19,45 @@ T = TypeVar("T")
 
 
 def read_config(
-    *paths: Union[str, Path]
+    *paths: Union[str, Path], **kwargs: Any
 ) -> Tuple[Dict[str, Any], Tuple[Path, ...]]:
-    parser = configparser.ConfigParser(allow_no_value=True, strict=False)
+    kwargs.setdefault("allow_no_value", True)
+    kwargs.setdefault("strict", False)
+    parser = configparser.ConfigParser(**kwargs)
+
     filenames = list(
         map(
             lambda p: p.resolve(),
             filter(
                 lambda p: p.is_file(),
-                map(lambda x: Path(x).expanduser(), paths)
-            )
-        )
+                map(lambda x: Path(x).expanduser(), paths),
+            ),
+        ),
+    )
+    config_paths = parser.read(filenames)
+
+    result: Dict[str, Union[str, Dict[str, str]]] = dict(
+        parser.items(parser.default_section, raw=True),
     )
-    paths = parser.read(filenames)
 
-    result = dict(parser.items(parser.default_section, raw=True))
     for section in parser.sections():
         config = dict(parser.items(section, raw=True))
         result[section] = config
 
-    return result, tuple(map(Path, paths))
+    return result, tuple(map(Path, config_paths))
 
 
-class ConfigAction(argparse.Action):
+class ConfigAction(Action):
     def __init__(
-        self, option_strings: Sequence[str], dest: str, search_paths=(),
-        type: Type[MappingProxyType] = MappingProxyType({}), help: str = '',
-        default: Any = None
+        self, option_strings: Sequence[str], dest: str,
+        search_paths: Iterable[Union[str, Path]] = (),
+        type: MappingProxyType = MappingProxyType({}), help: str = "",
+        default: Any = None,
     ):
         if not isinstance(type, MappingProxyType):
             raise ValueError("type must be MappingProxyType")
         super().__init__(
-            option_strings, dest, type=Path, help=help, default=default
+            option_strings, dest, type=Path, help=help, default=default,
         )
         self.search_paths = list(map(Path, search_paths))
         self._result = None
@@ -67,7 +75,6 @@ class Actions(str, Enum):
     APPEND = "append"
     APPEND_CONST = "append_const"
     COUNT = "count"
-    EXTEND = "extend"
     HELP = "help"
     PARSERS = "parsers"
     STORE = "store"
@@ -76,9 +83,12 @@ class Actions(str, Enum):
     STORE_TRUE = "store_true"
     VERSION = "version"
 
+    if sys.version_info >= (3, 8):
+        EXTEND = "extend"
+
     @classmethod
     def default(cls):
-        return cls.STORE
+        return None
 
 
 class Nargs(Enum):
@@ -104,7 +114,7 @@ def deep_getattr(name, attrs: Dict[str, Any], *bases: Type) -> Any:
 def merge_annotations(
     annotations: Dict[str, Any], *bases: Type
 ) -> Dict[str, Any]:
-    result = {}
+    result: Dict[str, Any] = {}
 
     for base in bases:
         result.update(getattr(base, "__annotations__", {}))
@@ -118,18 +128,26 @@ class StoreMeta(type):
             attrs.get("__annotations__", {}), *bases
         )
         attrs["__annotations__"] = annotations
-        attrs["_fields"] = tuple(annotations.keys())
+        attrs["_fields"] = tuple(
+            filter(
+                lambda x: not x.startswith("_"),
+                annotations.keys(),
+            ),
+        )
         return super().__new__(mcs, name, bases, attrs)
 
 
 class Store(metaclass=StoreMeta):
     _default_value = object()
+    _fields: Tuple[str, ...]
 
     def __new__(cls, **kwargs):
         obj = super().__new__(cls)
 
         type_map: Dict[str, Tuple[Type, Any]] = {}
         for key, value in obj.__annotations__.items():
+            if key.startswith("_"):
+                continue
             type_map[key] = (value, getattr(obj, key, cls._default_value))
 
         for key, (value_type, default) in type_map.items():
@@ -139,7 +157,7 @@ class Store(metaclass=StoreMeta):
             setattr(obj, key, value)
         return obj
 
-    def copy(self, **overrides) -> "Store":
+    def copy(self, **overrides):
         kwargs = self.as_dict()
         for key, value in overrides.items():
             kwargs[key] = value
@@ -152,16 +170,13 @@ class Store(metaclass=StoreMeta):
         }
 
     def __repr__(self) -> str:
-        values = ", ".join([f"{k!s}={v!r}" for k, v in self.as_dict().items()])
+        values = ", ".join([
+            f"{k!s}={v!r}" for k, v in sorted(self.as_dict().items())
+        ])
         return f"<{self.__class__.__name__}: {values}>"
 
 
 class ArgumentBase(Store):
-    aliases: Iterable[str] = frozenset()
-    nargs: Optional[Union[int, Nargs]] = Nargs.default()
-    action: Union[Actions, argparse.Action] = Actions.default()
-    type: Any = None
-
     def __init__(self, **kwargs):
         self._values = collections.OrderedDict()
 
@@ -199,22 +214,23 @@ class ArgumentBase(Store):
         return {k: v for k, v in kwargs.items() if v is not None}
 
 
-class Argument(ArgumentBase):
-    action: Union[Actions, argparse.Action] = Actions.default()
+class _Argument(ArgumentBase):
+    action: Union[Actions, Type[Action]] = Actions.default()
     aliases: Iterable[str] = frozenset()
-    choices: Optional[Iterable[T]] = None
-    const: Optional[T] = None
-    default: Optional[T] = None
+    choices: Optional[Iterable[str]] = None
+    const: Optional[Any] = None
+    default: Optional[Any] = None
+    env_var: Optional[str] = None
     help: Optional[str] = None
     metavar: Optional[str] = None
     nargs: Optional[Union[int, Nargs]] = Nargs.default()
     required: Optional[bool] = None
-    env_var: Optional[str] = None
     type: Any = None
 
 
-class Config(Argument):
-    action: ConfigAction = ConfigAction
+class _Config(_Argument):
+    """ Parse INI file and set results as a value """
+    action: Type[ConfigAction] = ConfigAction
     search_paths: Optional[Iterable[Union[Path, str]]] = None
 
 
@@ -239,9 +255,9 @@ class Meta(ABCMeta):
             try:
                 value = deep_getattr(key, attrs, *bases)
             except KeyError:
-                value = Argument(type=kind)
+                value = _Argument(type=kind)
 
-            if isinstance(value, Argument):
+            if isinstance(value, _Argument):
                 if value.type is None:
                     value.type = kind
                 arguments[key] = value
@@ -249,7 +265,7 @@ class Meta(ABCMeta):
                 argument_groups[key] = value
 
         for key, value in attrs.items():
-            if isinstance(value, Argument):
+            if isinstance(value, _Argument):
                 arguments[key] = value
             elif isinstance(value, AbstractGroup):
                 argument_groups[key] = value
@@ -264,28 +280,42 @@ class Meta(ABCMeta):
 
 
 class Base(metaclass=Meta):
-    __arguments__: Mapping[str, Argument]
+    __arguments__: Mapping[str, _Argument]
     __argument_groups__: Mapping[str, "Group"]
     __subparsers__: Mapping[str, "Parser"]
 
-    def __repr__(self) -> str:
-        name = self.__class__.__name__
-        attrs = {}
-        for attr in self.__arguments__.keys():
-            attrs[attr] = getattr(self, attr)
-        for attr in self.__argument_groups__.keys():
-            attrs[attr] = getattr(self, attr)
-        for attr in self.__subparsers__.keys():
-            attrs[attr] = getattr(self, attr)
+    def __getattribute__(self, item: str) -> Any:
+        value = super().__getattribute__(item)
+        if item.startswith("_"):
+            return value
 
-        values = ", ".join(f"{k}={v!r}" for k, v in attrs.items())
-        return f"{name}({values})"
+        if item in self.__arguments__:
+            class_value = getattr(self.__class__, item, None)
+            if value is class_value:
+                raise AttributeError(f"Attribute {item!r} was not parsed")
+        return value
+
+    def __repr__(self) -> str:
+        return (
+            f"<{self.__class__.__name__}: "
+            f"{len(self.__arguments__)} arguments, "
+            f"{len(self.__argument_groups__)} groups, "
+            f"{len(self.__subparsers__)} subparsers>"
+        )
+
+
+DestinationsType = MutableMapping[
+    str,
+    Set[Tuple[Base, str, Optional[_Argument], Optional[Action]]],
+]
 
 
 class Group(AbstractGroup, Base):
-    def __init__(self, title: str = None, description: Optional[str] = None,
-                 prefix: Optional[str] = None,
-                 defaults: Optional[Dict[str, Any]] = None):
+    def __init__(
+        self, title: str = None, description: Optional[str] = None,
+        prefix: Optional[str] = None,
+        defaults: Optional[Dict[str, Any]] = None,
+    ):
         self._prefix: Optional[str] = prefix
         self._title: Optional[str] = title
         self._description: Optional[str] = description
@@ -308,23 +338,24 @@ class Parser(AbstractParser, Base):
     )
 
     @staticmethod
-    def _add_argument(parser: Any, argument: Argument, dest: str, *aliases):
+    def _add_argument(
+        parser: Any, argument: _Argument, dest: str, *aliases,
+    ) -> Tuple[str, Action]:
         kwargs = argument.get_kwargs()
-        kwargs["dest"] = dest
 
-        if argument.is_positional:
-            dest = kwargs.pop("dest", None)
+        if not argument.is_positional:
+            kwargs["dest"] = dest
 
         if argument.default is not None:
-            kwargs['help'] = (
+            kwargs["help"] = (
                 f"{kwargs.get('help', '')} (default: {argument.default})"
             ).strip()
 
         if argument.env_var is not None:
-            kwargs['default'] = os.getenv(
-                argument.env_var, kwargs.get('default')
+            kwargs["default"] = os.getenv(
+                argument.env_var, kwargs.get("default"),
             )
-            kwargs['help'] = (
+            kwargs["help"] = (
                 f"{kwargs.get('help', '')} [ENV: {argument.env_var}]"
             ).strip()
 
@@ -334,54 +365,78 @@ class Parser(AbstractParser, Base):
     def get_cli_name(name: str) -> str:
         return name.replace("_", "-")
 
-    def get_env_var(self, name: str, argument: Argument) -> Optional[str]:
+    def get_env_var(self, name: str, argument: _Argument) -> Optional[str]:
         if argument.env_var is not None:
             return argument.env_var
         if self._auto_env_var_prefix is not None:
-            return f'{self._auto_env_var_prefix}{name}'.upper()
+            return f"{self._auto_env_var_prefix}{name}".upper()
         return None
 
     def __init__(
-        self, *args,
-        config_files: Iterable[Union[str, Path]] = (),
+        self, config_files: Iterable[Union[str, Path]] = (),
         auto_env_var_prefix: Optional[str] = None,
         **kwargs,
     ):
         super().__init__()
+        self.current_subparser = None
+        self._config_files = config_files
         self._config, filenames = read_config(*config_files)
-        epilog = kwargs.pop("epilog", "")
-        epilog += self.HELP_APPENDIX_PREAMBLE.format(
+
+        self._epilog = kwargs.pop("epilog", "")
+        self._epilog += self.HELP_APPENDIX_PREAMBLE.format(
             configs=repr(config_files),
         )
 
         if filenames:
-            epilog += self.HELP_APPENDIX_CURRENT.format(
+            self._epilog += self.HELP_APPENDIX_CURRENT.format(
                 num_existent=len(filenames),
-                existent=repr(list(map(str, filenames)))
+                existent=repr(list(map(str, filenames))),
             )
-        epilog += self.HELP_APPENDIX_END
+        self._epilog += self.HELP_APPENDIX_END
+        self._auto_env_var_prefix = auto_env_var_prefix
+        self._parser_kwargs = kwargs
 
-        self._parser = ArgumentParser(*args, epilog=epilog, **kwargs)
-        self._groups = {}
-        self._destinations = {}
-        self._auto_env_var_prefix: Optional[str] = auto_env_var_prefix
-        self._subparsers: Optional[argparse.ArgumentParser] = None
+    def _make_parser(
+        self, parser: Optional[ArgumentParser] = None,
+    ) -> Tuple[ArgumentParser, DestinationsType]:
+        if parser is None:
+            parser = ArgumentParser(
+                epilog=self._epilog, **self._parser_kwargs
+            )
 
+        destinations: DestinationsType = collections.defaultdict(set)
+
+        self._fill_arguments(destinations, parser)
+        self._fill_groups(destinations, parser)
+        if self.__subparsers__:
+            self._fill_subparsers(destinations, parser)
+
+        return parser, destinations
+
+    def _fill_arguments(
+        self, destinations: DestinationsType, parser: ArgumentParser,
+    ) -> None:
         for name, argument in self.__arguments__.items():
             aliases = set(argument.aliases)
+
+            # Add default alias
             if not aliases:
                 aliases.add(f"--{self.get_cli_name(name)}")
-            argument = argument.copy(
-                env_var=self.get_env_var(name, argument),
-                default=self._config.get(name, argument.default)
-            )
-            dest, action = self._add_argument(
-                self._parser, argument, name, *aliases
-            )
-            self._destinations[dest] = (self, name, argument, action)
 
+            argument = argument.copy(
+                aliases=aliases,
+                env_var=self.get_env_var(name, argument),
+                default=self._config.get(name, argument.default),
+            )
+
+            dest, action = self._add_argument(parser, argument, name, *aliases)
+            destinations[dest].add((self, name, argument, action))
+
+    def _fill_groups(
+        self, destinations: DestinationsType, parser: ArgumentParser,
+    ) -> None:
         for group_name, group in self.__argument_groups__.items():
-            parser = self._parser.add_argument_group(
+            group_parser = parser.add_argument_group(
                 title=group._title,
                 description=group._description,
             )
@@ -394,63 +449,154 @@ class Parser(AbstractParser, Base):
                 if not aliases:
                     aliases.add(f"--{self.get_cli_name(dest)}")
 
-                dest, action = self._add_argument(
-                    parser,
-                    argument.copy(
-                        default=config.get(
-                            name, group._defaults.get(name, argument.default)
-                        ),
-                        env_var=self.get_env_var(dest, argument),
+                argument = argument.copy(
+                    default=config.get(
+                        name, group._defaults.get(name, argument.default),
                     ),
-                    dest, *aliases
+                    env_var=self.get_env_var(dest, argument),
                 )
-                self._destinations[dest] = (group, name, argument, action)
+                dest, action = self._add_argument(
+                    group_parser, argument, dest, *aliases
+                )
+                destinations[dest].add((group, name, argument, action))
 
-        if self.__subparsers__:
-            raise NotImplementedError()
-
-    def _get_destinations(self, ns: argparse.Namespace) -> "_Destination":
-        for attr, dest_value in self._destinations.items():
-            target, name, argument, action = dest_value
-            value = getattr(ns, attr)
-            yield _Destination(
-                attr=attr, target=target, name=name,
-                argument=argument, action=action, value=value,
-            )
-
-    def parse_args(self, args: Optional[Sequence[str]] = None) -> "Parser":
-        ns: argparse.Namespace = self._parser.parse_args(args)
-
-        destinations: Iterable[_Destination] = list(self._get_destinations(ns))
-        configs = list(
-            filter(lambda x: x.argument.action == ConfigAction, destinations),
+    def _fill_subparsers(
+        self, destinations: DestinationsType, parser: ArgumentParser,
+    ) -> None:
+        subparsers = parser.add_subparsers()
+        subparser: Parser
+        destinations["current_subparser"].add(
+            (self, "current_subparser", None, None),
         )
 
-        for config_dest in configs:
-            config_dest.action(self._parser, ns, config_dest.value, None)
+        for subparser_name, subparser in self.__subparsers__.items():
+            current_parser, subparser_dests = (
+                subparser._make_parser(
+                    subparsers.add_parser(
+                        subparser_name, **subparser._parser_kwargs
+                    ),
+                )
+            )
+            current_parser.set_defaults(current_subparser=subparser)
+            for dest, values in subparser_dests.items():
+                for target, name, argument, action in values:
+                    destinations[dest].add((
+                        subparser, name, argument, action,
+                    ))
 
-        for dest in self._get_destinations(ns):
-            setattr(dest.target, dest.name, dest.value)
+    def parse_args(self, args: Optional[Sequence[str]] = None) -> "Parser":
+        parser, destinations = self._make_parser()
+        parsed_ns = parser.parse_args(args)
+
+        for key, values in destinations.items():
+            parsed_value = getattr(parsed_ns, key)
+            for target, name, argument, action in values:
+                if (
+                    target is not self and
+                    isinstance(target, Parser) and
+                    parsed_ns.current_subparser is not target
+                ):
+                    continue
+
+                if isinstance(action, ConfigAction):
+                    action(parser, parsed_ns, parsed_value, None)
+                    parsed_value = getattr(parsed_ns, key)
+                setattr(target, name, parsed_value)
 
         return self
 
     def print_help(self):
-        return self._parser.print_help()
+        parser, _ = self._make_parser()
+        return parser.print_help()
 
 
-class _Destination(NamedTuple):
-    attr: str
-    target: Parser
-    name: str
-    argument: Argument
-    action: Union[Actions, ConfigAction]
-    value: Any
+class ParserNamespace(SimpleNamespace):
+    def __init__(self, **kwargs: Any):
+        super().__init__(**kwargs)
+        self._destinations: Dict[str, Callable[[Any], Any]] = {}
+
+    @staticmethod
+    def on_value(obj: object, attr: str, value: Any) -> None:
+        setattr(obj, attr, value)
+
+    def add_destination(self, dest: str, obj: object, attr: str, setter=None):
+        if setter is None:
+            setter = partial(self.on_value, obj, attr)
+        self._destinations[dest] = setter
+
+    def __iter__(self):
+        return iter(self._destinations)
+
+    def __setattr__(self, key: str, value: Any):
+        if key.startswith("_"):
+            return super().__setattr__(key, value)
+        target = self._destinations[key]
+        target(value)
+        super(ParserNamespace, self).__setattr__(key, value)
 
 
-LogLevel = Argument(
-    choices=('debug', 'info', 'warning', 'error', 'critical'),
+# noinspection PyPep8Naming
+def Argument(
+    *aliases: str,
+    action: Union[Actions, Type[Action]] = Actions.default(),
+    choices: Optional[Iterable[str]] = None,
+    const: Optional[Any] = None,
+    default: Optional[Any] = None,
+    env_var: Optional[str] = None,
+    help: Optional[str] = None,
+    metavar: Optional[str] = None,
+    nargs: Optional[Union[int, str, Nargs]] = Nargs.default(),
+    required: Optional[bool] = None,
+    type: Any = None
+) -> Any:
+    return _Argument(
+        action=action,
+        aliases=aliases,
+        choices=choices,
+        const=const,
+        default=default,
+        env_var=env_var,
+        help=help,
+        metavar=metavar,
+        nargs=nargs,
+        required=required,
+        type=type,
+    )    # type: ignore
+
+
+# noinspection PyPep8Naming
+def Config(
+    *aliases: str,
+    search_paths: Optional[Iterable[Union[Path, str]]] = None,
+    choices: Optional[Iterable[str]] = None,
+    const: Optional[Any] = None,
+    default: Optional[Any] = None,
+    env_var: Optional[str] = None,
+    help: Optional[str] = None,
+    metavar: Optional[str] = None,
+    nargs: Optional[Union[int, str, Nargs]] = Nargs.default(),
+    required: Optional[bool] = None,
+    type: Any = None
+) -> Any:
+    return _Config(
+        search_paths=search_paths,
+        aliases=aliases,
+        choices=choices,
+        const=const,
+        default=default,
+        env_var=env_var,
+        help=help,
+        metavar=metavar,
+        nargs=nargs,
+        required=required,
+        type=type,
+    )    # type: ignore
+
+
+LogLevel: Any = Argument(
+    choices=("debug", "info", "warning", "error", "critical"),
     type=lambda v: getattr(logging, v.upper(), logging.INFO),
-    default='info'
+    default="info",
 )
 
 
